#include <stdio.h>
#include "LuaPlusCD.h"

extern "C" {
#include "lualib.h"
} // extern "C"

static int LS_LOG(lua_State* L)
{
    printf("In static function\n");
    return 0;
}


void LOG(const char* message)
{
    printf("In global function: %s\n", message);
}


class Logger
{
public:
    void LOGMEMBER(const char* message)
    {
        printf("In member function: %s\n", message);
    }

    virtual void LOGVIRTUAL(const char* message)
    {
        printf("In virtual member function: %s\n", message);
    }

	int LS_LOGMEMBER(lua_State* L)
    {
        printf("In member function.  Message: %s\n", lua_tostring(L, 1));
        return 0;
    }

    virtual int LS_LOGVIRTUAL(lua_State* L)
    {
        printf("In virtual member function\n");
        return 0;
    }
};

void Test1()
{
	lua_State* L = lua_open();

	lua_pushstring(L, "LOG");
	lua_pushfunctorclosure(L, LS_LOG, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_dostring(L, "LOG()");

	Logger logger;
	lua_pushstring(L, "LOGMEMBER");
	lua_pushfunctorclosure(L, logger, Logger::LS_LOGMEMBER, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_dostring(L, "LOGMEMBER('The message')");

	lua_pushstring(L, "LOGVIRTUAL");
	lua_pushfunctorclosure(L, logger, Logger::LS_LOGVIRTUAL, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_dostring(L, "LOGVIRTUAL()");


	lua_close(L);
}


float Add(float num1, float num2)
{
    return num1 + num2;
}


void AddTest()
{
	lua_State* L = lua_open();
	luaopen_base(L);

	lua_pushstring(L, "Add");
	lua_pushdirectclosure(L, Add, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_dostring(L, "print(Add(10, 5))");
	lua_dostring(L, "print(Add(10, 'Hello'))");  // Error!

	lua_close(L);
}


void DirectCallTest()
{
	lua_State* L = lua_open();
	luaopen_base(L);

	lua_pushstring(L, "LOG");
	lua_pushdirectclosure(L, LOG, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	Logger logger;
	lua_pushstring(L, "LOGMEMBER");
	lua_pushdirectclosure(L, logger, Logger::LOGMEMBER, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_pushstring(L, "LOGVIRTUAL");
	lua_pushdirectclosure(L, logger, Logger::LOGVIRTUAL, 0);
	lua_settable(L, LUA_GLOBALSINDEX);

	lua_dostring(L, "LOG('Hello')");
	lua_dostring(L, "LOGMEMBER('Hello')");
	lua_dostring(L, "LOGVIRTUAL('Hello')");

	lua_close(L);
}


class MultiObject
{
public:
	MultiObject(int num) :
		m_num(num)
	{
	}

	int Print(lua_State* state)
	{
		printf("%d\n", m_num);
		return 0;
	}

	void Print2(int num)
	{
		printf("%d %d\n", m_num, num);
	}

protected:
	int m_num;
};


void MultiObjectTest()
{
	lua_State* L = lua_open();
	luaopen_base(L);
	lua_pop(L, 1);

	// Create the metatable.
	lua_newtable(L);				// m
	lua_pushstring(L, "__index");	// m s
	lua_pushvalue(L, -2);			// m s m
	lua_settable(L, -3);			// m

	// Add the Print function.
	lua_pushstring(L, "Print");
	lua_pushobjectfunctorclosure(L, MultiObject::Print, 0);
	lua_settable(L, -3);

	// Add the Print2 direct function.
	lua_pushstring(L, "Print2");
	lua_pushobjectdirectclosure(L, (MultiObject*)0, MultiObject::Print2, 0);
	lua_settable(L, -3);

	MultiObject obj1(10);
	lua_pushstring(L, "obj1");		// m o
	lua_boxpointer(L, &obj1);		// m o p
	lua_pushvalue(L, -3);			// m o p m
	lua_setmetatable(L, -2);		// m o p
	lua_settable(L, LUA_GLOBALSINDEX);	// m

	MultiObject obj2(20);
	lua_pushstring(L, "obj2");		// m o
	lua_boxpointer(L, &obj2);		// m o p
	lua_pushvalue(L, -3);			// m o p m
	lua_setmetatable(L, -2);		// m o p
	lua_settable(L, LUA_GLOBALSINDEX);	// m

	lua_dostring(L, "obj1:Print()");
	lua_dostring(L, "obj2:Print()");
	lua_dostring(L, "obj1:Print2(5)");
	lua_dostring(L, "obj2:Print2(15)");

	lua_pushstring(L, "table1");	// m s
	lua_newtable(L);				// m s t
	lua_pushstring(L, "__object");	// m s t s
	lua_pushlightuserdata(L, &obj1);// m s t s p
	lua_settable(L, -3);			// m s t
	lua_pushvalue(L, -3);			// m s t m
	lua_setmetatable(L, -2);		// m s t
	lua_settable(L, LUA_GLOBALSINDEX);	// m

	lua_pushstring(L, "table2");	// m s
	lua_newtable(L);				// m s t
	lua_pushstring(L, "__object");	// m s t s
	lua_pushlightuserdata(L, &obj2);// m s t s p
	lua_settable(L, -3);			// m s t
	lua_pushvalue(L, -3);			// m s t m
	lua_setmetatable(L, -2);		// m s t
	lua_settable(L, LUA_GLOBALSINDEX);	// m

	lua_dostring(L, "table1:Print()");
	lua_dostring(L, "table2:Print()");
	lua_dostring(L, "table1:Print2(5)");
	lua_dostring(L, "table2:Print2(15)");

	lua_close(L);
}


main()
{
	Test1();
	AddTest();
	DirectCallTest();
	MultiObjectTest();
}
